/*
 * Copyright (c) 2011 NTT DATA Corporation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package jp.terasoluna.fw.batch.executor;

import jp.terasoluna.fw.batch.constants.EventConstants;
import jp.terasoluna.fw.batch.constants.JobStatusConstants;
import jp.terasoluna.fw.batch.constants.LogId;
import jp.terasoluna.fw.batch.executor.vo.BLogicResult;
import jp.terasoluna.fw.batch.executor.vo.BatchJobData;
import jp.terasoluna.fw.batch.util.JobUtil;
import jp.terasoluna.fw.dao.QueryDAO;
import jp.terasoluna.fw.dao.UpdateDAO;
import jp.terasoluna.fw.logger.TLogger;
import jp.terasoluna.fw.util.PropertyUtil;

import org.springframework.transaction.PlatformTransactionManager;
import org.springframework.transaction.TransactionStatus;
import org.springframework.transaction.support.DefaultTransactionDefinition;

/**
 * 񓯊ob`GO[L[^ۃNXB<br>
 * <br>
 * 񓯊WuNp̃ob`GO[L[^B
 * @see jp.terasoluna.fw.batch.executor.BatchExecutor
 * @see jp.terasoluna.fw.batch.executor.AbstractBatchExecutor
 * @see jp.terasoluna.fw.batch.executor.AsyncBatchExecutor
 */
public abstract class AbstractJobBatchExecutor extends AbstractBatchExecutor {

    /**
     * O.
     */
    private static final TLogger LOGGER = TLogger.getLogger(AbstractJobBatchExecutor.class);

    /**
     * Wu̎sԊui~bj擾pL[.
     */
    protected static final String JOB_INTERVAL_TIME = "polling.interval";

    /**
     * Executoȑ풓[h̏ItOĎt@CitpXŋLqj擾pL[.
     */
    protected static final String EXECUTOR_END_MONITORING_FILE = "executor.endMonitoringFile";

    /**
     * Executor̃WuI҂`FbNԊui~bj擾pL[.
     */
    protected static final String EXECUTOR_JOB_TERMINATE_WAIT_INTERVAL_TIME = "executor.jobTerminateWaitInterval";

    /**
     * Wu̎sԊui~bj̃ftHgl
     */
    protected static final long DEFAULT_JOB_INTERVAL_TIME = 1000;

    /**
     * WusGCsitrue/falsej̃ftHgl
     */
    protected static final boolean DEFAULT_JOB_AFTER_GC = true;

    /**
     * Executoȑ풓[h̃WuXg擾Ԋui~bj̃ftHgl
     */
    protected static final long DEFAULT_EXECUTOR_LOOP_INTERVAL_TIME = 1000;

    /**
     * Executoȑ풓[h̏ItOĎt@CitpXŋLqj̃ftHgl
     */
    protected static final String DEFAULT_EXECUTOR_END_MONITORING_FILE = null;

    /**
     * Executor̃XbhL[TCY`FbNԊui~bj̃ftHgl
     */
    protected static final long DEFAULT_EXECUTOR_QUEUE_CHECK_INTERVAL_TIME = 1000;

    /**
     * Executor̃WuI҂`FbNԊui~bj̃ftHgl
     */
    protected static final long DEFAULT_EXECUTOR_JOB_TERMINATE_WAIT_INTERVAL_TIME = 5000;

    /**
     * Wu̎sԊui~bj
     */
    protected long jobIntervalTime = DEFAULT_JOB_INTERVAL_TIME;

    /**
     * Executoȑ풓[h̏ItOĎt@CitpXŋLqj
     */
    protected String executorEndMonitoringFile = DEFAULT_EXECUTOR_END_MONITORING_FILE;

    /**
     * Executor̃WuI҂`FbNԊui~bj
     */
    protected long executorJobTerminateWaitIntervalTime = DEFAULT_EXECUTOR_JOB_TERMINATE_WAIT_INTERVAL_TIME;

    /**
     * JñXe[^XύXsǂ
     */
    protected boolean changeStartStatus = false;

    /**
     * RXgN^
     */
    protected AbstractJobBatchExecutor() {
        super();
        initParameter();
    }

    /*
     * (non-Javadoc)
     * @see jp.terasoluna.fw.batch.executor.AbstractBatchExecutor# initDefaultAppContext()
     */
    @Override
    protected void initDefaultAppContext() {
        // VXeAvP[VReLXg擾
        String defaultAppContextName = getDefaultBeanFileName();
        if (defaultAppContextName == null || "".equals(defaultAppContextName)) {
            LOGGER.error(LogId.EAL025003);
            return;
        }
        // f[^\[XAvP[VReLXg擾
        String dataSourceAppContextName = getDataSourceBeanFileName();
        if (dataSourceAppContextName == null
                || "".equals(dataSourceAppContextName)) {
            LOGGER.error(LogId.EAL025003);
            return;
        }

        defaultApplicationContext = getApplicationContext(
                defaultAppContextName, dataSourceAppContextName);
        if (defaultApplicationContext == null) {
            LOGGER.error(LogId.EAL025016, defaultAppContextName, dataSourceAppContextName);
            return;
        }
    }

    /*
     * (non-Javadoc)
     * @see jp.terasoluna.fw.batch.executor.AbstractBatchExecutor# initSystemDatasourceDao()
     */
    @Override
    protected void initSystemDatasourceDao() {
        if (defaultApplicationContext == null) {
            return;
        }

        String queryDaoKey = PropertyUtil
                .getProperty(SYSTEM_DATASOURCE_QUERY_DAO);
        String updateDaoKey = PropertyUtil
                .getProperty(SYSTEM_DATASOURCE_UPDATE_DAO);
        String transactionManagerKey = PropertyUtil
                .getProperty(SYSTEM_DATASOURCE_TRANSACTION_MANAGER);

        
        // QueryDAO擾
        if (queryDaoKey != null && queryDaoKey.length() != 0) {
            if (defaultApplicationContext.containsBean(queryDaoKey)) {
                try {
                    sysQueryDAO = (QueryDAO) defaultApplicationContext.getBean(
                            queryDaoKey, QueryDAO.class);
                } catch (Throwable e) {
                    LOGGER.error(LogId.EAL025017, e, queryDaoKey);
                }
            }
        }

        // UpdateDAO擾
        if (updateDaoKey != null && updateDaoKey.length() != 0) {
            if (defaultApplicationContext.containsBean(updateDaoKey)) {
                try {
                    sysUpdateDAO = (UpdateDAO) defaultApplicationContext
                            .getBean(updateDaoKey, UpdateDAO.class);
                } catch (Throwable e) {
                    LOGGER.error(LogId.EAL025018, e, updateDaoKey);
                }
            }
        }

        // transactionManager擾
        if (transactionManagerKey != null
                && transactionManagerKey.length() != 0) {
            if (defaultApplicationContext.containsBean(transactionManagerKey)) {
                try {
                    sysTransactionManager = (PlatformTransactionManager) defaultApplicationContext
                            .getBean(transactionManagerKey,
                                    PlatformTransactionManager.class);
                } catch (Throwable e) {
                    LOGGER.error(LogId.EAL025019, e, transactionManagerKey);
                }
            }
        }

        if (sysQueryDAO == null) {
            LOGGER.error(LogId.EAL025020);
        }
        if (sysUpdateDAO == null) {
            LOGGER.error(LogId.EAL025021);
        }
        if (sysTransactionManager == null) {
            LOGGER.error(LogId.EAL025022);
        }
    }

    /**
     * 
     */
    protected void initParameter() {
        // Wu̎sԊui~bj
        String jobIntervalTimeStr = PropertyUtil.getProperty(JOB_INTERVAL_TIME);
        if (jobIntervalTimeStr != null && jobIntervalTimeStr.length() != 0) {
            try {
                this.jobIntervalTime = Long.parseLong(jobIntervalTimeStr);
            } catch (NumberFormatException e) {
                this.jobIntervalTime = DEFAULT_JOB_INTERVAL_TIME;
            }
        } else {
            this.jobIntervalTime = DEFAULT_JOB_INTERVAL_TIME;
        }

        // Executoȑ풓[h̏ItOĎt@CitpXŋLqj
        String executorEndMonitoringFileStr = PropertyUtil
                .getProperty(EXECUTOR_END_MONITORING_FILE);
        if (executorEndMonitoringFileStr != null
                && executorEndMonitoringFileStr.length() != 0) {
            this.executorEndMonitoringFile = executorEndMonitoringFileStr;
        } else {
            this.executorEndMonitoringFile = DEFAULT_EXECUTOR_END_MONITORING_FILE;
        }

        // Executor̃WuI҂`FbNԊui~bj
        String executorJobTerminateWaitIntervalTimeStr = PropertyUtil
                .getProperty(EXECUTOR_JOB_TERMINATE_WAIT_INTERVAL_TIME);
        if (executorJobTerminateWaitIntervalTimeStr != null
                && executorJobTerminateWaitIntervalTimeStr.length() != 0) {
            try {
                this.executorJobTerminateWaitIntervalTime = Long
                        .parseLong(executorJobTerminateWaitIntervalTimeStr);
            } catch (NumberFormatException e) {
                this.executorJobTerminateWaitIntervalTime = DEFAULT_EXECUTOR_JOB_TERMINATE_WAIT_INTERVAL_TIME;
            }
        } else {
            this.executorJobTerminateWaitIntervalTime = DEFAULT_EXECUTOR_JOB_TERMINATE_WAIT_INTERVAL_TIME;
        }

    }

    /**
     * <h6>ob`s.</h6>
     * @param jobSequenceId WuV[PXR[h
     * @return rWlXWbNs
     */
    public BLogicResult executeBatch(String jobSequenceId) {
        BLogicResult result = new BLogicResult();
        boolean st = false;

        LOGGER.info(LogId.IAL025001, jobSequenceId);

        try {
            // SqlMapClientgp\`FbN
            if (sysQueryDAO == null || sysUpdateDAO == null
                    || sysTransactionManager == null) {
                LOGGER.error(LogId.EAL025023, jobSequenceId);
                return result;
            }

            // WuR[h擾
            BatchJobData jobRecord = JobUtil.selectJob(jobSequenceId, false,
                    sysQueryDAO);
            if (jobRecord == null) {
                LOGGER.error(LogId.EAL025024, jobSequenceId);
                return result;
            }

            if (changeStartStatus) {
                // WuXe[^XݒiJnj
                st = startBatchStatus(jobSequenceId, sysQueryDAO, sysUpdateDAO,
                        sysTransactionManager);
                if (!st) {
                    if (LOGGER.isInfoEnabled()) {
                        LOGGER.info(LogId.IAL025002, jobSequenceId, result.getBlogicStatus());
                    }
                    return result;
                }
            }

            // Ô߃g
            if (jobRecord.getJobAppCd() != null) {
                jobRecord.setJobAppCd(jobRecord.getJobAppCd().trim());
            }

            // ob`s
            result = executeBatch(jobRecord);

            if (LOGGER.isDebugEnabled()) {
                LOGGER.debug(LogId.DAL025021, result.getBlogicStatus());
            }
        } finally {
            // ς WuXe[^XݒiIj
            st = endBatchStatus(jobSequenceId, result, sysQueryDAO,
                    sysUpdateDAO, sysTransactionManager);

            if (!st) {
                if (LOGGER.isErrorEnabled()) {
                    LOGGER.error(LogId.EAL025025, jobSequenceId, result.getBlogicStatus());
                }
                return result;
            }
        }

        if (LOGGER.isInfoEnabled()) {
            LOGGER.info(LogId.IAL025003, jobSequenceId, result.getBlogicStatus());
        }
        return result;
    }

    /**
     * <h6>WuXe[^XXViWuJnj.</h6>
     * @param jobSequenceId XVΏۂ̃WuV[PXR[h
     * @param queryDAO QueryDAO
     * @param updateDAO UpdateDAO
     * @param transactionManager TransactionManager
     * @return WuXe[^XXVtrue
     */
    protected boolean startBatchStatus(String jobSequenceId, QueryDAO queryDAO,
            UpdateDAO updateDAO, PlatformTransactionManager transactionManager) {
        return updateBatchStatus(jobSequenceId,
                EventConstants.EVENT_STATUS_START, null, queryDAO, updateDAO,
                transactionManager);
    }

    /**
     * <h6>WuXe[^XXViWuIj.</h6>
     * @param jobSequenceId XVΏۂ̃WuV[PXR[h
     * @param result Xe[^X
     * @param queryDAO QueryDAO
     * @param updateDAO UpdateDAO
     * @param transactionManager TransactionManager
     * @return Xe[^XXVtrue
     */
    protected boolean endBatchStatus(String jobSequenceId, BLogicResult result,
            QueryDAO queryDAO, UpdateDAO updateDAO,
            PlatformTransactionManager transactionManager) {
        String blogicStatus = null;
        if (result != null) {
            blogicStatus = Integer.toString(result.getBlogicStatus());
        }
        return updateBatchStatus(jobSequenceId,
                EventConstants.EVENT_STATUS_NORMAL_TERMINATION, blogicStatus,
                queryDAO, updateDAO, transactionManager);
    }

    /**
     * <h6>WuXe[^XXV.</h6>
     * <p>
     * Xe[^X}bvɂāAWũXe[^X𔽉f
     * </p>
     * @param jobSequenceId XVΏۂ̃WuV[PXR[h
     * @param eventCode CxgR[h
     * @param blogicStatus blogic̖߂l
     * @param queryDAO QueryDAO
     * @param updateDAO UpdateDAO
     * @param transactionManager TransactionManager
     * @return Xe[^XXVtrue
     */
    protected boolean updateBatchStatus(String jobSequenceId, String eventCode,
            String blogicStatus, QueryDAO queryDAO, UpdateDAO updateDAO,
            PlatformTransactionManager transactionManager) {
        TransactionStatus tranStatus = null;

        try {
            DefaultTransactionDefinition tranDef = new DefaultTransactionDefinition();

            // gUNVJn
            tranStatus = transactionManager.getTransaction(tranDef);
            LOGGER.debug(LogId.DAL025022);

            // WuR[h擾
            BatchJobData job = JobUtil.selectJob(jobSequenceId, true, queryDAO);
            if (job == null) {
                LOGGER.error(LogId.EAL025026, jobSequenceId);
                return false;
            }

            // Xe[^X菈
            String changeStatus = judgmentStatus(job, jobSequenceId, eventCode,
                    blogicStatus);
            if (changeStatus == null) {
                return false;
            }

            LOGGER.debug(LogId.DAL025023, jobSequenceId, changeStatus);

            // Xe[^XXV
            JobUtil.updateJobStatus(job.getJobSequenceId(), changeStatus, null,
                    blogicStatus, queryDAO, updateDAO);

            // gUNVR~bg
            transactionManager.commit(tranStatus);
            LOGGER.debug(LogId.DAL025024);
        } catch (Exception e) {
            LOGGER.error(LogId.EAL025027, e);
            return false;
        } finally {
            try {
                // gUNVIigUNV̏ꍇ̓[obNj
                if (tranStatus != null && !tranStatus.isCompleted()) {
                    transactionManager.rollback(tranStatus);
                }
                LOGGER.debug(LogId.DAL025025);
            } catch (Exception e) {
                LOGGER.error(LogId.EAL025028, e);
            }
        }

        return true;
    }

    /**
     * <h6>WuXe[^X̍XV胁\bh</h6> CxgR[hƃWuXe[^XmFAWuXe[^X̍XVKvsB<br>
     * XVKvȂꍇInfoOo͂AnullԋpB
     * @param job WuR[h
     * @param jobSequenceId XVΏۂ̃WuV[PXR[h
     * @param eventCode CxgR[h
     * @param blogicStatus blogic̖߂l
     * @return
     */
    protected String judgmentStatus(BatchJobData job, String jobSequenceId,
            String eventCode, String blogicStatus) {

        String judge = null;

        if (EventConstants.EVENT_STATUS_START.equals(eventCode)) {
            if (JobStatusConstants.JOB_STATUS_UNEXECUTION.equals(job
                    .getCurAppStatus())) {
                judge = JobStatusConstants.JOB_STATUS_EXECUTING;
            } else {
                judge = null;
            }
        } else {
            if (JobStatusConstants.JOB_STATUS_EXECUTING.equals(job
                    .getCurAppStatus())) {
                judge = JobStatusConstants.JOB_STATUS_PROCESSED;
            } else {
                judge = null;
            }
        }

        // Xe[^XNG`FbN
        if (judge == null) {
            // INFOȂȂH
            if  (LOGGER.isInfoEnabled()) {
                LOGGER.info(LogId.IAL025004, jobSequenceId, blogicStatus, eventCode, job.getCurAppStatus(), judge);
            }
            return null;
        }

        return judge.toString();
    }

    /**
     * Wu̎sԊui~bj
     * @return the jobIntervalTime
     */
    public long getJobIntervalTime() {
        return jobIntervalTime;
    }

    /**
     * Executoȑ풓[h̏ItOĎt@CitpXŋLqj
     * @return the executorEndMonitoringFile
     */
    public String getExecutorEndMonitoringFile() {
        return executorEndMonitoringFile;
    }

    /**
     * Executor̃WuI҂`FbNԊui~bj
     * @return the executorJobTerminateWaitIntervalTime
     */
    public long getExecutorJobTerminateWaitIntervalTime() {
        return executorJobTerminateWaitIntervalTime;
    }

    /**
     * JñXe[^XύXsǂ
     * @param changeStartStatus the changeStartStatus to set
     */
    public void setChangeStartStatus(boolean changeStartStatus) {
        this.changeStartStatus = changeStartStatus;
    }

}
